import math
import torch
from torch.optim.optimizer import Optimizer
from torch.optim import AdamW


class AdamS(Optimizer):
    r"""
    Implements Adam with stable weight decay (AdamS) as proposed in
        "On the Overlooked Pitfalls of Weight Decay and How to Mitigate Them:
        A Gradient-Norm Perspective" (https://openreview.net/pdf?id=vnGcubtzR1).

    This implementation was from the git repo
        http://github.com/zeke-xie/stable-weight-decay-regularization/
            blob/master/swd_optim/adams.py (MIT license ca. July 2025)

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay coefficient (default: 1e-4)
    """

    def __init__(
        self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8, weight_decay=1e-4
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if not 0.0 <= weight_decay:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        defaults = {"lr": lr, "betas": betas, "eps": eps, "weight_decay": weight_decay}
        super().__init__(params, defaults)

    @torch.no_grad()
    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        param_size = 0
        exp_avg_sq_hat_sum = 0.0

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                param_size += p.numel()

                # Perform optimization step
                grad = p.grad
                if grad.is_sparse:
                    raise RuntimeError("AdamS does not support sparse gradients")

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(
                        p, memory_format=torch.preserve_format
                    )
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(
                        p, memory_format=torch.preserve_format
                    )

                beta1, beta2 = group["betas"]
                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]

                state["step"] += 1
                bias_correction2 = 1 - beta2 ** state["step"]

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)
                exp_avg_sq_hat = exp_avg_sq / bias_correction2

                exp_avg_sq_hat_sum += exp_avg_sq_hat.sum()

        # Calculate the sqrt of the mean of all elements in exp_avg_sq_hat
        exp_avg_mean_sqrt = math.sqrt(exp_avg_sq_hat_sum / param_size)

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue

                state = self.state[p]

                # Perform stable weight decay
                if group["weight_decay"] != 0:
                    p.data.mul_(
                        1 - group["weight_decay"] * group["lr"] / exp_avg_mean_sqrt
                    )

                beta1, beta2 = group["betas"]
                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                bias_correction1 = 1 - beta1 ** state["step"]
                bias_correction2 = 1 - beta2 ** state["step"]

                exp_avg_sq_hat = exp_avg_sq / bias_correction2

                denom = exp_avg_sq_hat.sqrt().add(group["eps"])

                step_size = group["lr"] / bias_correction1
                p.addcdiv_(exp_avg, denom, value=-step_size)

        # Make sure internal tensors are still leaf tensors
        # state['exp_avg'] = state['exp_avg'].detach()
        # state['exp_avg_sq'] = state['exp_avg_sq'].detach()

        return loss


def get_optimiser(model, optimiser=AdamW, lr=7e-4, weight_decay=5e-2):
    """
    Defaults are from one of the presets from the accompanying repo to Hassani
        et al. (2023) "Escaping the Big Data Paradigm with Compact Transformers",
        https://github.com/SHI-Labs/Compact-Transformers/blob/main/configs/
        pretrained/cct_7-3x1_cifar100_1500epochs.yml
    """
    weight_decay_exclude = []
    for keyword in [
        "bias",
        "norm",
        "embedding",
        "swiglu_beta",
        "sigma",
        "scale",
        "input_query",
        "reentrant_query",
    ]:
        weight_decay_exclude += [
            p for name, p in model.named_parameters() if keyword in name.lower()
        ]
    weight_decay_exclude = set(weight_decay_exclude)
    weight_decay_include = set(model.parameters()) - weight_decay_exclude
    return optimiser(
        [
            {"params": list(weight_decay_include)},
            {"params": list(weight_decay_exclude), "weight_decay": 0.0},
        ],
        weight_decay=weight_decay,
        lr=lr,
    )



"""
Based on Smith (2017) https://arxiv.org/abs/1506.01186
"""

from typing import Optional
import copy
import math

from scipy.ndimage import gaussian_filter1d
import scipy.constants as sc

from torch.amp import GradScaler

from .cycles import Cycle


class PASS:
    """
    A self-configuring learning rate scheduler
    """

    def __init__(
        self,
        learning_rate_schedule: Cycle,
        model,
        optimiser,
        scaler: Optional[GradScaler] = None,
        range_test: bool = True,
        max_lr: float = None,
        cool_point: float = None,
    ):
        if not range_test:
            assert max_lr is not None
            assert cool_point is not None

        self.model = model
        self.optimiser = optimiser
        self.scaler = scaler

        self.learning_rate_schedule = learning_rate_schedule

        self.range_test = range_test

        self.max_lr = max_lr
        self.cool_point = cool_point

        self.original_states = self._saved_states()

        self.range_test_results = []

        self.step_count = 0

        if range_test:
            self.start_range_test()  # sets LR to 1E-7

    @property
    def lr(self):
        """
        Return first lr from self.optimiser.param_groups
            (we assume they are all the same!)
        """
        for group in self.optimiser.param_groups:
            return group["lr"]

    @property
    def in_range_test(self):
        if not self.range_test:
            return False
        elif (len(self.range_test_results) == 0) or (
            not math.isnan(self.range_test_results[-1][1])
        ):
            return True
        else:
            return False

    @property
    def trained(self):
        if not self.range_test:
            return True
        elif math.isnan(self.range_test_results[-1][1]):
            return True
        else:
            return False

    @property
    def finished(self):
        return self.step_count >= len(self.learning_rate_schedule) - 1

    def _saved_states(self):
        saved_states = {
            "model": copy.deepcopy(self.model.state_dict()),
            "optimiser": copy.deepcopy(self.optimiser.state_dict()),
        }
        if self.scaler is not None:
            saved_states["scaler"] = copy.deepcopy(self.scaler.state_dict())
        return saved_states

    def save_states(self):
        self.saved_states = self._saved_states()

    def load_states(self, saved_states):
        self.model.load_state_dict(saved_states["model"])
        self.optimiser.load_state_dict(saved_states["optimiser"])
        if self.scaler is not None:
            self.scaler.load_state_dict(saved_states["scaler"])

    def recover_states(self):
        self.load_states(self.saved_states)

    @property
    def _schedule_lr(self):
        return (
            self.learning_rate_schedule(
                min(self.step_count, self.learning_rate_schedule.total_steps)
            )
            * (self.max_lr - self.cool_point)
            + self.cool_point
        )

    def set_lr(self, lr):
        for group in self.optimiser.param_groups:
            group["lr"] = lr

    def scale_lr(self, scaling_factor):
        self.set_lr(self.lr * scaling_factor)

    def start_range_test(self):
        self.save_states()
        self.optimiser.load_state_dict(self.original_states["optimiser"])
        if self.scaler is not None:
            self.scaler.load_state_dict(self.original_states["scaler"])
        self.set_lr(1e-7)

    def end_range_test(self):
        self.recover_states()
        self.update_learning_rates()

    def _smoothed_range_test(self, range_test_results):
        range_test_results = sorted(range_test_results, key=lambda x: x[0])
        learning_rates = [t[0] for t in range_test_results]
        losses = [t[1] for t in self.range_test_results]
        losses = losses[:-1] + [10 * max(losses)]
        smoothed_losses = gaussian_filter1d(losses, 3)
        return list(zip(learning_rates, smoothed_losses, strict=True))

    def _plot_range_test(self, range_test_results):
        """
        Returns a tuple with x values (learning rates) and y values (losses)
            which can then be passed to e.g. pyplot. We recommend presenting
            the plot with a logarithmic x axis.
        """
        range_test_results = sorted(range_test_results, key=lambda x: x[0])
        learning_rates = [t[0] for t in range_test_results]
        losses = [t[1] for t in range_test_results]
        return learning_rates, losses

    def _apply_range_test_result(self):
        """
        ...
        """
        range_test_results = self._smoothed_range_test(self.range_test_results)
        minimum = min(range_test_results, key=lambda x: x[1])
        points_left_of_min = [p for p in range_test_results if p[0] < minimum[0]]
        highest_point_left_of_min = max(points_left_of_min, key=lambda x: x[1])
        golden_min_loss = (highest_point_left_of_min[1] + minimum[1]) / (1 + sc.golden)
        golden_max_loss = (highest_point_left_of_min[1] + minimum[1]) / sc.golden
        for r in range_test_results:
            if r[1] < golden_max_loss:
                self.cool_point = r[0]
                break
        for r in range_test_results:
            if r[1] < golden_min_loss:
                self.max_lr = r[0]
            else:
                break
        print("High LR", self.max_lr)
        print("Cool point", self.cool_point)

    def update_learning_rates(self):
        if self.finished:
            pass
        else:
            self.set_lr(self._schedule_lr)

    def _append_to_range_test(self, loss_item: float):

        self.range_test_results.append((self.lr, loss_item))

        if math.isnan(loss_item):
            self._apply_range_test_result()
            self.end_range_test()
        else:
            # Continue range test, step up learning rate
            self.scale_lr(1.05)

    def step(self, loss_item: float):
        """
        This function manages the process of
            * Doing an initial range test
            * Training for one microcycle using the learning rates from the
                  initial range test ("burn in")
            * Doing a second range test to set the learning rate schedule for
                  the rest of training
            * Updating learning rates during training according to the macrocycle
        """
        if self.in_range_test:  # True at init unless self.range_test = False
            assert self.step_count == 0  # No weight updates yet
            self._append_to_range_test(loss_item)
        elif self.trained and not self.finished:
            self.step_count += 1
            self.update_learning_rates()
        else:
            pass
